/* CLParser Version 3. Copyright (c) Alan Ballard 1991. */

/*****************************************************************************
 *               CLParser Version 3  - md.h                                  *
 *                                                                           *
 *  Machine Dependent types and definitions.                                 *
 *                                                                           *
 *  This header file contains machine dependent definitions.  The types      *
 *  defined here are used to centralize dependencies on word size etc.       *
 *                                                                           *
 *  The file version.h should define a code to identify the compiler/machine *
 *  (unless such a code is predefined by the compiler, in which case         *
 *  version.h should be empty).                                              *
 *                                                                           *
 *  The first part of this file defines a number of attributes of the        *
 *  implementation.  This section consists of a number of #if ... #elif ...  *
 *  cases, based on the version code.                                        *
 *                                                                           *
 *  The second part defines a number of types and macros based on the        *
 *  attributes. This should generally not need changing to add a new version.*
 *                                                                           *
 *****************************************************************************/
    
/* Define a symbol that says this file has been included. */
#define MD_INCL

/* One of the following should be defined in the file version.h
 *    _MSC    - Microsoft C 5.1
 *    _QC     - Microsoft Quick C 2.0.  Predefined by compiler.
 *    _TURBOC - Turbo C 2.0
 *    _AIX370 - AIX/370 Metaware C compiler.  Predefined by compiler.
 *    _AIXPS2 - AIX for PS/2
 *    THINK_C - Think C for Mac (predefined)
 *    _MPW    - MPW C for Mac
 *    _AMIGA  - Lattice C for the Amiga    
 *    NeXT    - GNU C compiler for the NeXT. Predefined by compiler.
 */
#include "version.h"

/* Define attributes of the version.
 *
 * The following should be predefined in limits.h, according to ANSI.
 * If they are missing, add appropriate values here in the correct #elif case.
 *
 * INT_MAX       - max value of an int.
 * INT_MIN       - min value of an int.
 * UINT_MAX      - max value of an unsigned int.
 * SHRT_MAX      - max value of a short.
 * SHRT_MIN      - min value of a short.
 * USHRT_MAX     - max value of an unsigned short.
 * CHAR_BIT      - number of bits in a character.
 * 
 * The following should be #defined if required for the version, according
 * to the described conditions. They must be left undefined otherwise.
 *
 * BIGENDIAN      - define this if the target machine is "bigendian"; that is
 *                  if low-order (right-most) byte of an int has a higher
 *                  address than the high-order byte. 
 * STRICTALIGN    - define this if the target machine has a "strict alignment"
 *                  requirement; that is, if a word fetch from an odd-address
 *                  can cause a program interrupt. 
 * DOSOPTCHAR     - define this to allow command line options
 *                  to be specified as /option as well as -option.
 *
 * The following should be #defined to non-null values as described:
 *
 * REGISTERS      - define this as the number of register variables that it
 *                  is appropriate to use in the code.  The parser source 
 *                  attempts to prioritize its use of register variables
 *                  according to the setting of this.  For a good optimizing
 *                  compiler, it may be best to set this to 0, leaving all
 *                  register usage to the discretion of the compiler.
 *                  This will default to 0 if not specified.
 */   
 
#include <limits.h>

#if defined(_MSC) || defined(_QC) || defined(_TURBOC)

   /* Various compilers for PC DOS or OS/2. */

   #undef BIGENDIAN                 /* 80x86 is little endian */
   #undef STRICTALIGN               /* 80x86 has no alignment requirement */
   #define REGISTERS 0              /* MSC compiler knows best */
   #define DOSOPTCHAR               /* Allow DOS option flag. */

#elif defined(_AIXPS2)
  
   /* PS/2 AIX compiler. */

   #undef BIGENDIAN                 /* 80386 is little endian */
   #undef STRICTALIGN               /* 80386 has no alignment requirement */
   #define REGISTERS 0              /* compiler knows best */

#elif defined(_AIX370)

   /* 370 AIX Metaware compiler. */

   #define BIGENDIAN                /* 370 is bigendian */
   #undef STRICTALIGN               /* 370 has no alignment requirement */
   #define REGISTERS 3              /* a few register variables are OK */

#elif defined(_MPW) || defined(THINK_C)

   /* Macintosh MPW and Think C (68000) */
   
   #define BIGENDIAN
   #define STRICTALIGN              /* for 68000 */
   #define REGISTERS 3
      
#elif defined(_AMIGA)

   /* Lattice C for the Amiga */
   
   #define BIGENDIAN                /* 680x0 is bigendian */
   #define STRICTALIGN              /* for 68000 */
   #define REGISTERS 0              /* compiler knows best */

#elif defined(NeXT)

   /* GNU C for the NeXT */
   
   #define BIGENDIAN                /* 680x0 is bigendian */
   #undef STRICTALIGN               /* 68030 or 68040 doesn't need it */
   #define REGISTERS 0              /* let the compiler do it */
   #undef USHRT_MAX                 /* NeXT botched the definition of these, */
   #undef UINT_MAX                  /* so we redefine them here */
   #define USHRT_MAX 0xFFFF
   #define UINT_MAX 0xFFFFFFFF
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Types and macros are defined based on the preceding attributes. */

/* Integer sizes.  Define INT for a 16bit (or bigger if necessary) int,
 * and UINT for unsigned version. */


#if INT_MAX == 32767 && INT_MIN <= -32767
   typedef int INT;
      
#elif SHRT_MAX >= 32767 && SHRT_MIN <= -32767
   typedef short INT;

#else
   typedef int INT;                  /* int may be bigger than wanted. */

#endif

#if UINT_MAX == 65535
   typedef unsigned int UINT;
      
#elif USHRT_MAX >= 65535
   typedef unsigned short UINT;

#else
   typedef unsigned int UINT;        /* maybe bigger than 16 bits. */

#endif

/* Some of the data in the syntax tables consists of two byte integers
 * packed without regard to any alignment boundaries that might
 * be present on the target system, and with the high-order byte first.
 * When accessing such data, the macro FETCH16 is always used to assign it
 * to an INT.  The first parameter of this is the destination INT.
 * The second parameter will an address of the two bytes.  This may be
 * a void * or a BYTE *, so must be cast appropriately.
 * The FETCH16 macro must be coded as appropriate for the specific
 * machine/compiler environment. If the machine has no alignment concerns,
 * and bytes are ordered within INT's as in the table, a simple assignment
 * of the form
 *       (dest = *(INT *)(source))
 * is all that is required.
 * If there is an alignment requirement or byte reversal is required,
 * or if BYTEs aren't exactly 8 bits, the INT result must be built by 
 * fetching the bytes and shifting/or'ing.
 *
 * The macro STORE16 is similarly used by the table generator to build
 * these tables.  The first parameter is a pointer to the array element
 * where the INT is to be stored; the second parameter is the INT.
 * For byte-reversal or strict alignment requirements, the INT is
 * stored as two separate BYTE assignments.
 */

#if CHAR_BIT == 8 && !defined(STRICTALIGN) && defined(BIGENDIAN)
   #define FETCH16(dest, source) (dest = *(INT *)(source))
   #define STORE16(dest, source) (*(INT *)(dest) = source)
      
#else      
   #define FETCH16(dest, source) (dest = ((*(BYTE *)(source)) << 8) \
             | *((BYTE *)(source)+1))
   #define STORE16(dest, source) (((*(BYTE *)(dest)) = (BYTE)((source) >> 8)),\
             (*((BYTE *)(dest)+1) = (BYTE)((source) & 0x00ff)))

#endif                

/* Define symbols used to control what variables get given the register
 * attribute.  Each of the symbols REGISTER1 through REGISTER5 is defined as 
 * either "register" or as null.
 */

#if !defined(REGISTERS)
   #define REGISTERS 0
#endif

#if REGISTERS >= 1
   #define REGISTER1 register
#else
   #define REGISTER1
#endif
#if REGISTERS >= 2
   #define REGISTER2 register 
#else
   #define REGISTER2
#endif
#if REGISTERS >= 3
   #define REGISTER3 register
#else
   #define REGISTER3
#endif
#if REGISTERS >= 4
   #define REGISTER4 register
#else
   #define REGISTER4
#endif
#if REGISTERS >= 5
   #define REGISTER5 register
#else
   #define REGISTER5
#endif
