#include <Types.h>
#include <ctype.h>
#include <Memory.h>
#include <StdLib.h>
#include <stdIO.h>
#include <string.h>
#include <Connections.h>
#include <Packages.h>
#include <Dialogs.h>
#include "dlgutil.h"
#include "glob.h"
#include "log.h"
#include "clp.h"
#include "ctb.h"
#include "util.h"
#include "script_semantics.h"
#define CTBErr globals->ParseRC

#include "script_grammar.h"

/* Local procedures */
BOOL newFirstParm(PScript_Stack_Frame thisStackFrame, HVariable thisValue);
BOOL convertToString(HPCB Pcb, HVariable thisValue);
BOOL convertToInteger(HPCB Pcb, HVariable thisValue);
HVariable GetVariable(Script_Globals * globals, char *varName);
BOOL MakeLocalVariable(PScript_Stack_Frame thisStackFrame, char *varName);
HVariable NewVariable(char *varName, HVariable varList);
void DisposeVariables(HVariable thisVar);
BOOL ScriptTR(UINT Script_Error,short Start,short End);
#define globals ((Script_Globals *)gl)

BOOL Script_Stack(HPCB Pcb, void * gl, UINT Action);
BOOL Script_End(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Test(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Logical(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Strings(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Numbers(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Variables(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Arithmetic(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Error(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Function(HPCB Pcb, void * gl, UINT Action);
BOOL Script_String_Values(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Numeric_Values(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Null_Values(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Send_Functions(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Dialog_Functions(HPCB Pcb, void * gl, UINT Action);
BOOL Script_String_Functions(HPCB Pcb, void * gl, UINT Action);
BOOL Script_Get_Functions(HPCB Pcb, void * gl, UINT Action);

/* This procedure handles semantic actions for the interpreter */
BOOL Script_Semantics(HPCB Pcb, void * gl, UINT Action)
{
	BOOL (*Semantic_Function)(HPCB Pcb, void * gl, UINT Action)=NULL;
	
	switch (Action) {
	case Push:
	case Pop:
	case Pop_Clean:
		Semantic_Function = Script_Stack;
		break;
	case The_End:
		Semantic_Function = Script_End;
		break;
	case Switch_Test:
	case Switch_Off:
	case Test_While:
	case Do_Test:
		Semantic_Function = Script_Test;
		break;
	case Missing_Comment_End:
	case Missing_Right_Paren:
	case Missing_Closing_Quote:
	case Statement_Expected:
	case Invalid_String_Escape:
	case Value_Expected:
	case Variable_Name_Expected:
	case Missing_Right_Brace:
	case Invalid_Function:
		Semantic_Function = Script_Error;
	}
	
	if (Semantic_Function==NULL) {
		if (!globals->stack->doCommands) return (!gCancel);
		switch (Action) {
		case Less_Than:
		case Greater_Than:
		case Not_Equal:
		case Equal:
			Semantic_Function = Script_Logical;
			break;
		case Init_String:
		case String_Seg1:
		case String_Seg2:
		case String_Seg3:
		case String_Octal:
			Semantic_Function = Script_Strings;
			break;
		case Push_Integer:
			Semantic_Function = Script_Numbers;
			break;
		case Push_Variable:
		case Assign_Variable:
		case Var_Name:
		case Local_Variable:
		case Function_Name:
			Semantic_Function = Script_Variables;
			break;
		case Add:
		case Subtract:
		case Multiply:
		case Divide:
		case And:
		case Or:
			Semantic_Function = Script_Arithmetic;
			break;
		case Function:
			Semantic_Function = Script_Function;
		}
	}
	
	if (gCancel) return false;
	return (*Semantic_Function)(Pcb, gl, Action);
}


/* This procedure handles pushing and poping stack frames for the interpreter */

BOOL Script_Stack(HPCB Pcb, void * gl, UINT Action)
{	
	short thisPosition;
	register PScript_Stack_Frame thisFrame;
	
	Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
	
	if (Action == Push) {
	
		/* Get a new stack frame */
		thisFrame = (PScript_Stack_Frame)NewPtrClear(sizeof(Script_Stack_Frame));
		if (!thisFrame) return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		
		/* Put it at the top of the stack */
		thisFrame->nextFrame = globals->stack;
		globals->stack = thisFrame;
		
		/* Initialize things */
		thisFrame->doCommands = thisFrame->nextFrame->doCommands;
		thisFrame->parsePosition = thisPosition;
	}
	
	else { /* must be a stack pop */
	
		thisFrame = globals->stack;
		
		if (globals->TraceRefN && 
				(thisFrame->doCommands != thisFrame->nextFrame->doCommands)) {
			if (thisFrame->doCommands) fprintf(stdout, "Execution suspended.....\n");
			else fprintf(stdout, "Execution resumed.....\n");
		}
			
		/* Remove this frame from the stack */
		globals->stack = thisFrame->nextFrame;
		thisFrame->nextFrame = nil;
		
		/* Before we dispose of this level of the stack, we might have to 
		   pass a value up one level. If we are actually doing the commands
		   and there is a value to pass up, then do it. */
			   
		if (Action==Pop && thisFrame->doCommands) {
			HVariable nextParameter;
			if (!thisFrame->firstParameter) 
				return ScriptTR(Novalue_Function,thisFrame->parsePosition,thisPosition);
			nextParameter = (*thisFrame->firstParameter)->nextVariable;
			
			(*thisFrame->firstParameter)->nextVariable = globals->stack->firstParameter;
			globals->stack->firstParameter = thisFrame->firstParameter;
			thisFrame->firstParameter = nextParameter;
		}

		FreeStack(thisFrame);
	}
	
	return true;
}

/*
	This procedure makes sure we have parsed the whole script.
*/
BOOL Script_End(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Action)
	short Current_Position;
	Parse_Get(Pcb, Scan_Position, &Current_Position, 
		sizeof(Current_Position), false);
	return Current_Position == globals->Script_Length;
}

/*
	This procedure handles setting the execution flag for ifs and whiles.
*/

BOOL Script_Test(HPCB Pcb, void * gl, UINT Action)
{
	register PScript_Stack_Frame thisFrame = globals->stack;

	if (Action == Switch_Test) {
		/* We are in the first part of an if. If we are executing it then turn off
		   execution for the rest of the elseifs etc. Otherwise set the execution
		   flag to match what was going on one level back.
		*/
		if (thisFrame->doCommands || thisFrame->nextFrame->doCommands) {
			thisFrame->doCommands = !thisFrame->doCommands;
			if (globals->TraceRefN)
				if (thisFrame->doCommands) fprintf(stdout, "Execution resumed.....\n");
				else fprintf(stdout, "Execution suspended.....\n");
		}
	}
	
	/* The following items are only done if we are executing commands. */
	else if (thisFrame->doCommands) {
	
		switch (Action) {
		
		case Switch_Off:
			/* We have been executing an elseif, so turn off doCommands one level back
			   to stop execution of any following elseifs or a trailing else. */
			thisFrame->nextFrame->doCommands = false;
			break;
			
		case Test_While:
			/* Reset the parse pointer to the beginning of the while loop */
			Parse_Set(Pcb, Scan_Position, thisFrame->parsePosition-5);
			break;
			
		case Do_Test: {
			/* Test the expression in an if or while and set doCommands */
			if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
			thisFrame->doCommands = 
				(*thisFrame->firstParameter)->Value.Number != 0;
			if (globals->TraceRefN && !thisFrame->doCommands)
				fprintf(stdout, "Execution suspended.....\n");
		}
		}
	}
	
	return true;
}

/* 
	This routine does a comparison of two items. If the two items are numeric,
	then a numeric test is done. Otherwise the items are converted to strings
	and compared using strcmp for less than and greater than and strstr for 
	equal to and not equal to.
*/

BOOL Script_Logical(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)

	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable arg1, arg2;
	HVariable oldFirst;
		
	HLock((Handle)(thisFrame->firstParameter));
	arg2 = *(thisFrame->firstParameter);
	HLock((Handle)(arg2->nextVariable));
	arg1 = *(arg2->nextVariable);

	if (arg1->variableType == numericVariable &&
		arg2->variableType == numericVariable) {
		if (globals->TraceRefN) {
			fprintf(stdout, "Test: Left side = %ld.\n",arg1->Value.Number);
			fprintf(stdout, "Test: Right side = %ld.\n",arg2->Value.Number);
		}
		arg1->Value.Number =
			(Action==Less_Than)    ? arg1->Value.Number <  arg2->Value.Number  :
			(Action==Greater_Than) ? arg1->Value.Number >  arg2->Value.Number  :
			(Action==Not_Equal)    ? arg1->Value.Number != arg2->Value.Number  :
		 /* (Action==Equal) */		 arg1->Value.Number == arg2->Value.Number;
	}
	else {
		long Answer;
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		if (!convertToString(Pcb, arg2->nextVariable)) return false;
		if (globals->TraceRefN) {
			fprintf(stdout, "Test: Left side = \"%s\".\n",*arg1->Value.String);
			fprintf(stdout, "Test: Right side = \"%s\".\n",*arg2->Value.String);
		}
		Answer =
			(Action==Less_Than)    
				? strcmp(*arg1->Value.String, *arg2->Value.String) <  0 :
			(Action==Greater_Than) 
				? strcmp(*arg1->Value.String, *arg2->Value.String) >  0 :
			(Action==Not_Equal)    
				? strstr(*arg1->Value.String, *arg2->Value.String) == 0 :
		 /* (Action==Equal) */
				  strstr(*arg1->Value.String, *arg2->Value.String) != 0;
		DisposHandle(arg1->Value.String);
		arg1->Value.Number = Answer;
		arg1->variableType = numericVariable;
	}
	HUnlock((Handle)(arg2->nextVariable));
	
	oldFirst = thisFrame->firstParameter;
	thisFrame->firstParameter = arg2->nextVariable;
	arg2->nextVariable = nil;
	DisposeVariables(oldFirst);
		
	return true;
		
}

/* This routine builds up character strings. */

BOOL Script_Strings(HPCB Pcb, void * gl, UINT Action)
{
	register PScript_Stack_Frame thisFrame = globals->stack;

	/* Allocate a new (empty) variable to build the string constant in */
		
	if (Action == Init_String) {
		register PVariable newParm;
		
		if (!newFirstParm(thisFrame, NewVariable("", nil))) return false;
		newParm = *(thisFrame->firstParameter);
		
		newParm->variableType = stringVariable;
		if ( !(newParm->Value.String = NewHandleClear(1)) )
			return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		HUnlock((Handle)(thisFrame->lastParameter));
	}
	else {
		Handle ourString = (*thisFrame->firstParameter)->Value.String;
		int Size_Increase = 1;
		char Dummy;
		char *nextChar;
		
		/* All string segments are one character long except for String_Seg1.
		   We call Parse_Get to see how long the String_Seg1 segment is */
		   
		if (Action == String_Seg1)
			Size_Increase = Parse_Get(Pcb, Terminal_Text, &Dummy, 1, false)-1;
	
		/* Streach our handle the right amount and lock it down */
		
		SetHandleSize(ourString, GetHandleSize(ourString)+Size_Increase);
		if (MyMemErr())
			return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		HLock(ourString);
		nextChar = *ourString+strlen(*ourString);
		
		switch (Action) {
		case String_Seg1:
			/* Parse_Get will return the segment and a trailing nil */
			Parse_Get(Pcb, Terminal_Text, nextChar, Size_Increase+1, false);
			break;
		case String_Seg2:
			/* This segment is just an embeded double quote */
			*nextChar = '"';
			*++nextChar = 0;
			break;
		case String_Seg3:
			/* This segment is an escaped character. If it is N,R,T or B then
			   we have to translate it to the right thing. Note that Parse_Get
			   will put the trailing nil on for us. */
			Parse_Get(Pcb, Terminal_Text, nextChar, 2, true);
			*nextChar = (*nextChar == 'N') ? '\x0A' :
						(*nextChar == 'R') ? '\x0D' :
						(*nextChar == 'T') ? '\x09' :
						(*nextChar == 'B') ? '\x08' :
						*nextChar;
			break;
		case String_Octal: {
			/* This is an octal representation of a character. Decode it. */
			char Octal[4];
			char *nextOctal;
			Parse_Get(Pcb, Terminal_Text, Octal, 4, false);
			for (nextOctal = Octal; *nextOctal; nextOctal++)
				*nextChar = ((*nextChar)*8) + *nextOctal - '0';
			*++nextChar = 0;
			break;
		}
		}
		HUnlock(ourString);
	}
	return true;
}

/* This routine pushes a number onto the stack. */

BOOL Script_Numbers(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Action)

	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable newParm;
	char intString[20];
	
	if (!newFirstParm(thisFrame, NewVariable("", nil))) return false;
	newParm = *(thisFrame->firstParameter);
	
	Parse_Get(Pcb, Current_Production_Text, intString, sizeof(intString), false);
	newParm->variableType = numericVariable;
	newParm->Value.Number = atoi(intString);
	
	HUnlock((Handle)(thisFrame->lastParameter));
	return true;
}

/* This routine handles variables. */

BOOL Script_Variables(HPCB Pcb, void * gl, UINT Action)
{
	register PScript_Stack_Frame thisFrame = globals->stack;
	
	switch (Action) {
	case Var_Name:
	case Function_Name:
	case Local_Variable: {
		char Dummy;
		char *varName;
		/* Use Parse_Get to get the length of the variable name. */
		int varNameLength = Parse_Get(Pcb, Current_Production_Text, &Dummy, 1, true);
		if ( !(varName = NewPtr(varNameLength)) ) 
			return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		
		/* Now get the name itself */
		Parse_Get(Pcb, Current_Production_Text, varName, varNameLength, true);
		if (Action == Var_Name) {
			if ( !(thisFrame->tempVariable = GetVariable(globals, varName)) )
				return false;
		}
		else if (Action == Function_Name) {
			thisFrame->localVariables = NewVariable(varName, nil);
		}
		else
			if (!MakeLocalVariable(thisFrame, varName)) return false;
		DisposPtr(varName);
		break;
	}
	
	/* Push the value of a variable onto the stack */
	case Push_Variable: {
		register PVariable newParm;
	
		if (!newFirstParm(thisFrame, NewVariable("", nil))) return false;
		newParm = *(thisFrame->firstParameter);

		newParm->variableType = (*thisFrame->tempVariable)->variableType;
		newParm->Value.Number = (*thisFrame->tempVariable)->Value.Number;
		if (newParm->variableType == stringVariable)
			HandToHand(&newParm->Value.String);
		HUnlock((Handle)(thisFrame->firstParameter));
		break;
	}
	
	/* Assign the value on the stack to a variable */
	case Assign_Variable: {
		register PVariable arg1, res;
		HLock((Handle)(thisFrame->firstParameter));
		HLock((Handle)(thisFrame->tempVariable));
		arg1=*(thisFrame->firstParameter);
		res=*(thisFrame->tempVariable);
		if (res->variableType == stringVariable)
			DisposHandle(res->Value.String);
		res->variableType = arg1->variableType;
		res->Value.Number = arg1->Value.Number;
		arg1->Value.String = nil;	
		if (globals->TraceRefN)
			if (res->variableType == stringVariable)
				fprintf(stdout, "Assign variable %s = \"%s\".\n",
					res->variableName, *res->Value.String);
			else
				fprintf(stdout, "Assign variable %s = %ld.\n",
					res->variableName, res->Value.Number);
		DisposeVariables(thisFrame->firstParameter);
		thisFrame->firstParameter = nil;
		HUnlock((Handle)(thisFrame->tempVariable));
		break;
	}
	}
	return true;
}

/* This routine handles arithmetic functions */

BOOL Script_Arithmetic(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)

	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable arg1, arg2;
	HVariable oldFirst;
	
	if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
	arg2 = *(thisFrame->firstParameter);
	if (!convertToInteger(Pcb, arg2->nextVariable)) return false;
	arg1 = *(arg2->nextVariable);
	
	arg1->Value.Number =
		(Action==Add)      ? arg1->Value.Number + arg2->Value.Number :
		(Action==Subtract) ? arg1->Value.Number - arg2->Value.Number :
		(Action==Multiply) ? arg1->Value.Number * arg2->Value.Number :
		(Action==Divide)   ? arg1->Value.Number / arg2->Value.Number :
		(Action==Or)       ? arg1->Value.Number || arg2->Value.Number:
	 /* (Action==And) */	 arg1->Value.Number && arg2->Value.Number;
	 
	oldFirst = thisFrame->firstParameter;
	thisFrame->firstParameter = arg2->nextVariable;
	arg2->nextVariable = nil;
	DisposeVariables(oldFirst);

	return true;		
}

BOOL Script_Error(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb,gl)
	short thisPosition;
	
	/* A parsing error occurred. Call ScriptTR to report it. */
	
	if (CTBErr == ParseQuit || gCancel) return false;
	
	Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
	return ScriptTR(Action,globals->stack->parsePosition,thisPosition);
}

BOOL Script_Function(HPCB Pcb, void * gl, UINT Action)
{
	register BOOL (*Semantic_Function)(HPCB Pcb, void * gl, UINT Action)=NULL;
	register char *funcname=(*(globals->stack->localVariables))->variableName;
	register char *funcnamecomplete;
	
	switch (*funcname++) {
	case 'A':
		funcnamecomplete = "PNAME";
		Semantic_Function = Script_String_Values;
		Action = ApName_Function;
		break;
	case 'B':
		funcnamecomplete = "REAK";
		Semantic_Function = Script_Null_Values;
		Action = Break_Function;
		break;
	case 'C':
		switch (*funcname++) {
		case 'O':
			funcnamecomplete = "NCAT";
			Semantic_Function = Script_String_Functions;
			Action = Concat_Function;
			break;
		case 'T':
			funcnamecomplete = "BTOOL";
			Semantic_Function = Script_String_Values;
			Action = CTBTool_Function;
			break;
		}
		break;
	case 'D':
		if ('I'==(*funcname++)) {
			if ('A'==(*funcname++)) {
				if ('L'==(*funcname++)) {
					switch (*funcname++) {
					case 'O':
						funcnamecomplete = "G";
						Semantic_Function = Script_Dialog_Functions;
						Action = Dialog_Function;
						break;
					case 'U':
						if ('P'==(*funcname++)) {
							switch (*funcname++) {
							case 'P':
								funcnamecomplete = "ASSWORD";
								Semantic_Function = Script_String_Values;
								Action = DialupPassword_Function;
								break;
							case 'U':
								funcnamecomplete = "SERNAME";
								Semantic_Function = Script_String_Values;
								Action = DialupUsername_Function;
							}
						}
					}
				}
			}
		}
		break;
	case 'E':
		switch (*funcname++) {
		case 'C':
			funcnamecomplete = "HO";
			Semantic_Function = Script_Null_Values;
			Action = Echo_Function;
			break;
		case 'X':
			funcnamecomplete = "ISTS";
			Semantic_Function = Script_Numeric_Values;
			Action = Exists_Function;
			break;
		}
		break;
	case 'F':
		funcnamecomplete = "ORGETPASSWORD";
		Semantic_Function = Script_Null_Values;
		Action = ForgetPassword_Function;
		break;
	case 'G':
		if ('E'==(*funcname++)) {
			if ('T'==(*funcname++)) {
				if (0==(*funcname)) {
					funcnamecomplete = "";
					Semantic_Function = Script_Get_Functions;
					Action = Get_Function;
				}
				else {
					funcnamecomplete = "CONFIG";
					Semantic_Function = Script_String_Values;
					Action = GetConfig_Function;
				}
			}
		}
		break;
	case 'L':
		switch (*funcname++) {
		case 'E':
			funcnamecomplete = "NGTH";
			Semantic_Function = Script_Numeric_Values;
			Action = Length_Function;
			break;
		case 'I':
			funcnamecomplete = "NEEND";
			Semantic_Function = Script_String_Values;
			Action = LineEnd_Function;
		}
		break;
	case 'M':
		funcnamecomplete = "OUSEDOWN";
		Semantic_Function = Script_Numeric_Values;
		Action = MouseDown_Function;
		break;
	case 'P':
		switch (*funcname++) {
		case 'A':
			funcnamecomplete = "SSWORD";
			Semantic_Function = Script_String_Values;
			Action = Password_Function;
			break;
		case 'O':
			if ('P'==(*funcname++)) {
				if ('A'==(*funcname)) {
					funcnamecomplete = "ACCOUNT";
					Semantic_Function = Script_String_Values;
					Action = POPaccount_Function;
				}
				else {
					funcnamecomplete = "SERVER";
					Semantic_Function = Script_String_Values;
					Action = POPserver_Function;
				}
			}
			break;
		case 'R':
			funcnamecomplete = "OGRESS";
			Semantic_Function = Script_Send_Functions;
			Action = Progress_Function;
			break;
		case 'W':
			funcnamecomplete = "DIALOG";
			Semantic_Function = Script_Dialog_Functions;
			Action = Password_Dialog_Function;
		}
		break;
	case 'Q':
		if ('U'==(*funcname++)) {
			if ('I'==(*funcname++)) {
				if ('E'==(*funcname)) {
					funcnamecomplete = "ETSEND";
					Semantic_Function = Script_Send_Functions;
					Action = Quiet_Send_Function;
				}
				else {
					funcnamecomplete = "T";
					Semantic_Function = Script_Null_Values;
					Action = Quit_Function;
				}
			}
		}
		break;
	case 'R':
		funcnamecomplete = "ESTORE";
		Semantic_Function = Script_String_Values;
		Action = Restore_Function;
		break;
	case 'S':
		switch (*funcname++) {
		case 'A':
			if ('V'==(*funcname++)) {
				if ('E'==(*funcname++)) {
					if (0==(*funcname)) {
						funcnamecomplete = "";
						Semantic_Function = Script_Null_Values;
						Action = Save_Function;
					}
					else {
						funcnamecomplete = "PASSWORD";
						Semantic_Function = Script_Numeric_Values;
						Action = SavePassword_Function;
					}
				}
			}
			break;
		case 'E':
			funcnamecomplete = "ND";
			Semantic_Function = Script_Send_Functions;
			Action = Send_Function;
			break;
		case 'M':
			funcnamecomplete = "TPSERVER";
			Semantic_Function = Script_String_Values;
			Action = SMTPserver_Function;
			break;
		case 'U':
			funcnamecomplete = "BSTRING";
			Semantic_Function = Script_String_Functions;
			Action = Substring_Function;
		}
		break;
	case 'T':
		switch (*funcname++) {
		case 'E':
			funcnamecomplete = "LNETCOMMAND";
			Semantic_Function = Script_String_Values;
			Action = TelnetCommand_Function;
			break;
		case 'O':
			switch (*funcname++) {
			case 'L':
				funcnamecomplete = "OWER";
				Semantic_Function = Script_String_Values;
				Action = Tolower_Function;
				break;
			case 'U':
				funcnamecomplete = "PPER";
				Semantic_Function = Script_String_Values;
				Action = Toupper_Function;
			}
			break;
		case 'R':
			if ('A'==(*funcname++)) {
				if ('C'==(*funcname)) {
					funcnamecomplete = "CE";
					Semantic_Function = Script_Null_Values;
					Action = Trace_Function;
				}
				else {
					funcnamecomplete = "PECHO";
					Semantic_Function = Script_Null_Values;
					Action = Trapecho_Function;
				}
			}
		}
	}
	
	if (Semantic_Function == NULL) return false;
	if (strcmp(funcname,funcnamecomplete)) return false;
	return (*Semantic_Function)(Pcb, globals, Action);
}

/* This routine handles functions that return string values */

BOOL Script_String_Values(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)

	register PScript_Stack_Frame thisFrame = globals->stack;
	char newValue[256];
	register PVariable result;

	switch (Action) {
	case ApName_Function: {
		strcpy(newValue, "NewsWatcher 2.0d17");
		c2pstr(newValue);
		break;
	}
	case Tolower_Function: {
		char *nextChar;
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		strcpy(newValue, *(*thisFrame->firstParameter)->Value.String);
		for (nextChar = newValue; *nextChar; nextChar++) *nextChar = tolower(*nextChar);
		c2pstr(newValue);
		break;
	}
	case Toupper_Function: {
		char *nextChar;
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		strcpy(newValue, *(*thisFrame->firstParameter)->Value.String);
		for (nextChar = newValue; *nextChar; nextChar++) *nextChar = toupper(*nextChar);
		c2pstr(newValue);
		break;
	}
	case TelnetCommand_Function: {
		StringHandle ourString = GetString(1202);
		if (ourString) {
			strncpy(newValue, (char*)*ourString, *ourString[0]+1);
			if (thisFrame->firstParameter) {
				c2pstr(*((*(thisFrame->firstParameter))->Value.String));
				SetHandleSize((Handle)ourString, (*((*(thisFrame->firstParameter))->Value.String))[0]+1);
				strncpy((char*)*ourString, (char*)*((*(thisFrame->firstParameter))->Value.String), (*((*(thisFrame->firstParameter))->Value.String))[0]+1);
				ChangedResource((Handle)ourString);
				WriteResource((Handle)ourString);
			}
		}
		else newValue[0]=0;
		break;
	}
	case LineEnd_Function: {
		StringHandle ourString = GetString(1201);
		if (ourString) {
			strncpy(newValue, (char*)*ourString, *ourString[0]+1);
			if (thisFrame->firstParameter) {
				c2pstr(*((*(thisFrame->firstParameter))->Value.String));
				SetHandleSize((Handle)ourString, (*((*(thisFrame->firstParameter))->Value.String))[0]+1);
				strncpy((char*)*ourString, (char*)*((*(thisFrame->firstParameter))->Value.String), (*((*(thisFrame->firstParameter))->Value.String))[0]+1);
				ChangedResource((Handle)ourString);
				WriteResource((Handle)ourString);
			}
		}
		else newValue[0]=0;
		break;
	}
	case DialupUsername_Function: {
		StringHandle ourString = GetString(1103);
		if (ourString) strncpy(newValue, (char*)*ourString, *ourString[0]+1);
		else newValue[0]=0;
	}
	case DialupPassword_Function: {
		StringHandle ourString = GetString(1104);
		if (ourString) strncpy(newValue, (char*)*ourString, *ourString[0]+1);
		else {
			ResType ourResType;
			Str255 ourResName;
			short ourResNumber;
			char password[32];
			char userName[128];
			DialogPtr myDialog;
			Handle ourResource = GetNamedResource('DLOG',"\pPassword");
			{
				StringHandle ourString = GetString(1103);
				if (ourString) strncpy(userName, (char*)*ourString, *ourString[0]+1);
				else userName[0]=0;
			}
			ParamText((StringPtr)userName,"\pEnter","\p","\p");
			GetResInfo(ourResource, &ourResNumber, &ourResType, ourResName);
			myDialog = MyGetNewDialog(ourResNumber);
			password[0]=0;
			DlgSetCString(myDialog, 4, "");
			SetItemPassword(myDialog, 4, password);
			SetItemMaxLength(myDialog, 4, 31);
			SetItemFlash(myDialog, 5, true);
			for (;;) {
				short itemHit;
				MyModalDialog((ModalFilterProcPtr)DialogFilter, &itemHit, true, true);
				if (itemHit == ok) strcpy(newValue, password);
				if (itemHit == ok || itemHit == cancel) break;
			}
			MyDisposDialog(myDialog);
			if (!(*newValue)) return false;
			c2pstr(newValue);
			strncpy((char*)*ourString, newValue, newValue[0]+1);
			ChangedResource((Handle)ourString);
			WriteResource((Handle)ourString);
		}
		break;
	}
	case Password_Function: {
		if (!(*gPrefs.remotePassword)) {
			ResType ourResType;
			Str255 ourResName;
			short ourResNumber;
			char password[32];
			char userName[128];
			DialogPtr myDialog;
			Handle ourResource = GetNamedResource('DLOG',"\pPassword");
			strcpy(userName,gPrefs.name);
			c2pstr(userName);
			ParamText((StringPtr)userName,"\pEnter","\p","\p");
			GetResInfo(ourResource, &ourResNumber, &ourResType, ourResName);
			myDialog = MyGetNewDialog(ourResNumber);
			password[0]=0;
			DlgSetCString(myDialog, 4, "");
			SetItemPassword(myDialog, 4, password);
			SetItemMaxLength(myDialog, 4, 31);
			SetItemFlash(myDialog, 5, true);
			for (;;) {
				short itemHit;
				MyModalDialog((ModalFilterProcPtr)DialogFilter, &itemHit, true, true);
				if (itemHit == ok) strcpy(gPrefs.remotePassword, password);
				if (itemHit == ok || itemHit == cancel) break;
			}
			MyDisposDialog(myDialog);
			if (!(*gPrefs.remotePassword)) return false;
		}
		strcpy(newValue, gPrefs.remotePassword);
		c2pstr(newValue);
		break;
	}
	case POPaccount_Function: {
		strcpy(newValue, gPrefs.name);
		c2pstr(newValue);
		break;
	}
	case POPserver_Function: {
		strcpy(newValue, gPrefs.host);
		c2pstr(newValue);
		break;
	}
	case SMTPserver_Function:
		pstrcpy((StringPtr)newValue, (StringPtr)gPrefs.mailServerName);
		break;
	case CTBTool_Function: {
		extern ConnHandle gCnH;
		CMGetToolName((*gCnH)->procID, (StringPtr)newValue);
		break;
	}
	case GetConfig_Function: {
		char *CTB_Config, *Beginning, *End;
		extern ConnHandle gCnH;
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		CTB_Config = CMGetConfig(gCnH);
		if (!CTB_Config) return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		Beginning = strstr(CTB_Config, 
			*(*thisFrame->firstParameter)->Value.String);
		if (Beginning) {
			Beginning += strlen(*(*thisFrame->firstParameter)->Value.String) + 1;
			if (Beginning[0] == '"') End = strchr(Beginning = Beginning+1, '"');
			else End = strchr(Beginning, ' ');
			strncpy(newValue, Beginning, End - Beginning);
			newValue[End-Beginning] = 0;
			c2pstr(newValue);
		}
		else newValue[0] = 0;
		DisposPtr((Ptr)CTB_Config);
		break;
		}
	case Restore_Function:
		if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
		if (!(*thisFrame->firstParameter)->nextVariable) {
			StringHandle ourString = GetString((*thisFrame->firstParameter)->Value.Number);
			if (ourString) strncpy(newValue, (char*)*ourString, *ourString[0]+1);
			else {
				(*thisFrame->firstParameter)->variableType = undefinedVariable;
				return (!gCancel);
			}
			break;
		}
		if (!convertToInteger(Pcb, (*thisFrame->firstParameter)->nextVariable)) return false;
		GetIndString((StringPtr)newValue, (*thisFrame->firstParameter)->Value.Number,
			(*(*thisFrame->firstParameter)->nextVariable)->Value.Number);
		break;
	}

	if (!thisFrame->firstParameter) 
		if (!newFirstParm(thisFrame, NewVariable("", nil))) return false;
	HLock((Handle)(thisFrame->firstParameter));
	result = *(thisFrame->firstParameter);
	if (result->variableType==stringVariable) DisposHandle(result->Value.String);
	result->variableType = stringVariable;
	if ( !(result->Value.String = MyNewHandle(newValue[0]+1)) )
		return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
	strcpy(*result->Value.String, p2cstr((StringPtr)newValue));
	HUnlock((Handle)(thisFrame->firstParameter));

	return true;	
}

/* This routine handles functions that return numeric values */

BOOL Script_Numeric_Values(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)

	register PScript_Stack_Frame thisFrame = globals->stack;
	register long Answer;
	register PVariable result;
	
	switch (Action) {
	case Exists_Function:
		Answer = (undefinedVariable!=(*thisFrame->firstParameter)->variableType);
		break;
	case MouseDown_Function:
		Answer = Button();
		break;
	case Length_Function:
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		Answer = strlen(*(*thisFrame->firstParameter)->Value.String);
		break;
	case SavePassword_Function:
		Answer = gPrefs.savePassword;
		break;
	}
	
	if (!thisFrame->firstParameter) 
		if (!newFirstParm(thisFrame, NewVariable("", nil))) return false;
	HLock((Handle)(thisFrame->firstParameter));
	result = *(thisFrame->firstParameter);
	if (result->variableType==stringVariable) DisposHandle(result->Value.String);
	result->variableType = numericVariable;
	result->Value.Number = Answer;
	HUnlock((Handle)(thisFrame->firstParameter));

	return true;
}

/* This routine handles functions that don't return any value */

BOOL Script_Null_Values(HPCB Pcb, void * gl, UINT Action)
{
	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable arg1, arg2, arg3;
	
	if (Action == Quit_Function) {
		CTBErr = ParseQuit;
		if (globals->TraceRefN) fprintf(stdout, "Quit\n");
		return false;
	}
	
	if (Action == ForgetPassword_Function) {
		*gPrefs.remotePassword = 0;
	}
	
	else if (Action == Save_Function) {
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		arg1 = *(thisFrame->firstParameter);
		if (!convertToInteger(Pcb, arg1->nextVariable)) return false;
		arg2 = *(arg1->nextVariable);
		
		if ( !arg2->nextVariable ) {
			if (globals->TraceRefN) fprintf(stdout, "Save STR(%ld) = \"%s\"\n",
				arg2->Value.Number, *arg1->Value.String);
			c2pstr(*arg1->Value.String);
			{
				Handle resH = GetResource('STR ',arg2->Value.Number);
				if (resH) {
					SetHandleSize(resH, (*arg1->Value.String)[0]+1);
					strncpy(*resH, *arg1->Value.String, (*arg1->Value.String)[0]+1);
					ChangedResource(resH);
					WriteResource(resH);
				}
			}
		}
		
		else {
		
			if (!convertToInteger(Pcb, arg2->nextVariable)) return false;
			arg3 = *(arg2->nextVariable);
			if (globals->TraceRefN) fprintf(stdout, "Save STR#(%ld,%ld) = \"%s\"\n",
				arg3->Value.Number, arg2->Value.Number, *arg1->Value.String);
			c2pstr(*arg1->Value.String);
			/* ChangeStrn(arg3->Value.Number, arg2->Value.Number,
				*arg1->Value.String); */
			WriteResource(GetResource('STR#', arg3->Value.Number));
		}
	}

	else {

		if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
		arg1 = *(thisFrame->firstParameter);
		
		switch (Action) {
	
		case Echo_Function:
			globals->Echo = arg1->Value.Number;
			break;
		case Trapecho_Function:
			globals->TrapEcho = arg1->Value.Number;
			break;
		case Break_Function: {
			Str255 data;
			sprintf((Ptr)data, "BREAK [%d ticks]", arg1->Value.Number);
			if (globals->Echo) StatusWindow((Ptr)data);
			if (globals->TraceRefN) fprintf(stdout, "Break (%ld)\n", arg1->Value.Number);
			CTBErr = CTBBreak(arg1->Value.Number);
			if (CTBErr) return false;
			break;
			}
		case Trace_Function: {
			int temp = arg1->Value.Number;
			Parse_Set(Pcb, Parse_Trace, temp);
			Create("\pTrace", 0, '????', 'TEXT');
			freopen("Trace", "w+", stdout);
			globals->TraceRefN = true;
			break;
		}
		}
	}
	
	DisposeVariables(thisFrame->firstParameter);
	thisFrame->firstParameter=nil;
 
 	return true;
}

/* The various send functions */

BOOL Script_Send_Functions(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)

	register PScript_Stack_Frame thisFrame = globals->stack;
	char *ourString;
		
	/* Concatenate all the parameters */
	if (!Script_String_Functions(Pcb, globals, Concat_Function)) return false;

	HLock((Handle)((*thisFrame->firstParameter)->Value.String));
	ourString = *(*thisFrame->firstParameter)->Value.String;
	
	if (Action == Progress_Function || globals->Echo) {
		if (Action != Quiet_Send_Function) {
			StatusWindow(ourString);
		}
		else {
			StatusWindow("secret things");
		}
	}
			
	if (globals->TraceRefN) fprintf(stdout, "Text sent -> \"%s\"\n", ourString);
	if (Action != Progress_Function) {
		/* if (gPrefs.logActionsToFile) {
			LogNavigateSent(ourString, strlen(ourString));
		} */
		if (CTBErr=CTBSendTrans(1, ourString, strlen(ourString))) return false;
	}

	DisposeVariables(thisFrame->firstParameter);
	thisFrame->firstParameter=nil;

	return true;
}

/* The dialog functions */

BOOL Script_Dialog_Functions(HPCB Pcb, void * gl, UINT Action)
{
	register PScript_Stack_Frame thisFrame = globals->stack;
	short thisPosition;
	short itemhit, ourResNumber;
	DialogPtr myDialog;
	register PVariable arg1;
	register HVariable nextArg;
	char password[32];
	Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
		
	if (!thisFrame->firstParameter) 
		return ScriptTR(Missing_Parameter,thisFrame->parsePosition,thisPosition);

	/* Loop through up to 4 more parameters and use them as arguments to ParamText */
	nextArg = thisFrame->firstParameter;
	{
		StringPtr param[4] = {"\p", "\p", "\p", "\p"};
		int i;
		for (i=0; i<=3;i++) {
			if (!(*nextArg)->nextVariable) break;
			if (!convertToString(Pcb, nextArg)) return false;
			arg1 = *(nextArg);
			param[i] = c2pstr(*arg1->Value.String);
			nextArg = arg1->nextVariable;
		}
		ParamText(param[(i+3)&3],param[(i+2)&3],param[(i+1)&3],param[i&3]);
	}

	if ((*nextArg)->nextVariable)
		return ScriptTR(Missing_Parameter,thisFrame->parsePosition,thisPosition);

	HLock((Handle)nextArg);
	arg1 = *(nextArg);
	
	if (arg1->variableType == undefinedVariable) 
		return ScriptTR(Uninitialized_Variable,thisFrame->parsePosition,thisPosition);
		
	/* If the argument is a string, use it to look for a DLOG resource with that
	   name. Otherwise use the numeric value as a resource number */
	if (arg1->variableType == stringVariable) {
		ResType ourResType;
		Str255 ourResName;
		Handle ourResource;
		if (!(ourResource = GetNamedResource('DLOG', c2pstr(*arg1->Value.String))))
			return ScriptTR(Invalid_Resource_Name,thisFrame->parsePosition,thisPosition);
		GetResInfo(ourResource, &ourResNumber, &ourResType, ourResName);
	}
	else
		ourResNumber = arg1->Value.Number;
			
	/* if (!MommyMommy(ATTENTION,nil)) return false; */
	
	/* Put up the dialog */
	
	if (!(myDialog = MyGetNewDialog(ourResNumber))) 
		return ScriptTR(Invalid_Resource_Number,thisFrame->parsePosition,thisPosition);
	ShowWindow(myDialog);
		
	/* Initialize the text fields */
	{
		int i;
		BOOL select_field = false;
		for (i=1;;i++) {
			short itemType;
			Handle item;
			Rect box;
			itemType=-1;
			GetDItem(myDialog, i, &itemType, &item, &box);
			if (itemType == -1) break;
			if ((itemType==editText) || (itemType==editText+itemDisable)) {
				HVariable stringvar;
				char varName[15];
				sprintf(varName, "DIALOGSTR%d", i);
				if ( !(stringvar=GetVariable(globals, varName)) )
					return false;
				if ((*stringvar)->variableType == undefinedVariable) {
					Str255 theText;
					(*stringvar)->variableType = stringVariable;
					GetIText(item,theText);
					p2cstr(theText);
					if ( !((*stringvar)->Value.String=NewHandle(strlen((char *)theText)+1)) )
						return ScriptTR(Memory_Error,MemError(), __LINE__);
					strcpy(*(*stringvar)->Value.String, (char *)theText);
				}
				{
					if (!convertToString(Pcb, stringvar)) return false;
					if (Action == Password_Dialog_Function && i != 4) {
						strcpy(password, *(*stringvar)->Value.String);
						memset(*(*stringvar)->Value.String, '', strlen(password));
					}
					SetIText(item, c2pstr(*(*stringvar)->Value.String));
				}
				if (!select_field) {
					SelIText(myDialog, i, 0, 32767);
					select_field = true;
				}
			}
		}
	}
	
	/* Do the appropriate things until the user hits a button */
	
	if (Action == Password_Dialog_Function) {
		SetItemPassword(myDialog, 4, password);
		SetItemMaxLength(myDialog, 4, 31);
		SetItemFlash(myDialog, 5, true);
	}

	for (;;) {
		short itemType;
		Handle item;
		Rect box;
		MyModalDialog((ModalFilterProcPtr)DialogFilter, &itemhit, true, true);
		GetDItem(myDialog, itemhit, &itemType, &item, &box);
		if (itemType != editText) break;
	}
	
	/* Return the item number of the button hit */
	
	arg1 = *(thisFrame->firstParameter);
	if (arg1->variableType==stringVariable) DisposHandle(arg1->Value.String);
	arg1->variableType = numericVariable;
	arg1->Value.Number = itemhit;
	DisposeVariables(arg1->nextVariable);
	arg1->nextVariable = nil;
	
	/* Load the text fields into variables */
	
	{
		int i;
		for (i=1; ;i++) {
			short itemType;
			Handle item;
			Rect box;
			char theText[256];
			itemType = -1;
			GetDItem(myDialog, i, &itemType, &item, &box);
			if (itemType == -1) break;
			if ((itemType == editText) || (itemType == editText+itemDisable)) {
				HVariable stringvar;
				char varName[15];
				sprintf(varName, "DIALOGSTR%d", i); 
				if ( !(stringvar = GetVariable(globals, varName)) ) 
					return false;
				(*stringvar)->variableType = stringVariable;
				if (Action == Dialog_Function || i != 4) {
					GetIText(item, (StringPtr)theText);
					p2cstr((StringPtr)theText);
				}
				else
					strcpy(theText, password);
				if ( !((*stringvar)->Value.String=MyNewHandle(strlen(theText)+1)) )
					return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
				strcpy(*(*stringvar)->Value.String, theText);
				if (globals->TraceRefN)
					fprintf(stdout, "Assign variable %s = \"%s\".\n",
						varName,theText);
				HUnlock((Handle)stringvar);
			}
		}
	}
	MyDisposDialog(myDialog);

	return true;
}

/* The various string functions. */

BOOL Script_String_Functions(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb)
	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable arg1, arg2, arg3;
	HVariable oldFirst;
	
	switch (Action) {

	case Substring_Function:
		if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
		arg2 = *(thisFrame->firstParameter);
		
		if (!arg2->nextVariable) {
			short thisPosition;
			Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
			return ScriptTR(Missing_Parameter, thisFrame->parsePosition, thisPosition);
		}
		arg1 = *(arg2->nextVariable);
		
		if (!arg1->nextVariable) arg3 = nil;
		else {
			if (!convertToInteger(Pcb, arg2->nextVariable)) return false;
			arg3 = arg2;
			arg2 = arg1;
			arg1 = *(arg2->nextVariable);
		}
		if (!convertToString(Pcb, arg2->nextVariable)) return false;
		
		if (arg2->Value.Number > strlen(*arg1->Value.String))
			(*arg1->Value.String)[0] = 0;
		else if (arg2->Value.Number > 0)
			strcpy(*arg1->Value.String, *arg1->Value.String+arg2->Value.Number);
		if (arg3 && arg3->Value.Number < strlen(*arg1->Value.String)) {
			(*arg1->Value.String)[arg3->Value.Number] = 0;
		}
		SetHandleSize(arg1->Value.String, strlen(*arg1->Value.String)+1);
		
		oldFirst = thisFrame->firstParameter;
		thisFrame->firstParameter = arg2->nextVariable;
		arg2->nextVariable = nil;
		DisposeVariables(oldFirst);
		
		break;
		
	case Concat_Function:
		if (!convertToString(Pcb, thisFrame->firstParameter)) return false;
		arg1 = *(thisFrame->firstParameter);
		arg2 = arg1;
		while (arg2->nextVariable) {
			if (!convertToString(Pcb, arg2->nextVariable)) return false;
			arg2 = *(arg2->nextVariable);
			HUnlock((Handle)(arg2->Value.String));
			SetHandleSize(arg2->Value.String, 
				strlen(*arg1->Value.String)+strlen(*arg2->Value.String)+1);
			if (MyMemErr())
				return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
			strcat(*arg2->Value.String, *arg1->Value.String);
			DisposHandle(arg1->Value.String);
			arg1->Value.String = arg2->Value.String;
			arg2->Value.String = nil;
		}
		break;
	}
	HUnlock((Handle)(arg1->Value.String));
	HUnlock((Handle)(thisFrame->firstParameter));

	return true;
}

/* The Get function */

BOOL Script_Get_Functions(HPCB Pcb, void * gl, UINT Action)
{
#pragma unused(Pcb,Action)

	register PScript_Stack_Frame thisFrame = globals->stack;
	register PVariable arg1;
	char inputString[256];
	long bSize;
	
	CTBErr=noErr;
	
	if (!convertToInteger(Pcb, thisFrame->firstParameter)) return false;
	arg1 = *(thisFrame->firstParameter);
	
	/* Set the timeout to the value of the argument (or something long if zero) */
	gCTBTimeout = arg1->Value.Number;
	if (!gCTBTimeout) gCTBTimeout = 45;
	
	/* Prepare the top of the stack to return a string */
	arg1->variableType = stringVariable;
	if ( !(arg1->Value.String=NewHandleClear(1)) )
		return ScriptTR(Memory_Error,MyMemErr(),__LINE__);

	bSize = sizeof(inputString)-1;
	CTBErr=NetRecvLine(inputString+1, &bSize);
		/* if (gPrefs.logActionsToFile) {
			LogNavigateResp((Ptr)inputString+1, bSize);
		} */
	if ((!CTBErr || CTBErr==cmTimeOut) && bSize) {
		/* Assign gCancel to ParseRC and return if true */
		if (globals->ParseRC = gCancel) return false;
		inputString[0] = bSize;
		p2cstr((StringPtr)inputString);
		if (globals->Echo) StatusWindow(inputString);
		/* Assign the received string to the top of the stack */
		SetHandleSize(arg1->Value.String, strlen(inputString)+1);
		if (MyMemErr())
			return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		HLock((Handle)(arg1->Value.String));
		strcpy(*arg1->Value.String, inputString);
	}

	if (CTBErr && (CTBErr != cmTimeOut)) return false;
	CTBErr = noErr;
	if (globals->TraceRefN) fprintf(stdout, "Text returned -> \"%s\"\n", 
		*arg1->Value.String);
	HUnlock((Handle)(arg1->Value.String));
	HUnlock((Handle)(thisFrame->firstParameter));

	return true;
		
}

/* Stick a new element at the top of the stack and lock it */
BOOL newFirstParm(PScript_Stack_Frame thisStackFrame, HVariable thisValue)
{
	/* If we were passed a nil handle, we can assume a memory error occured */
	if (!thisValue) return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
	HLock((Handle)thisValue);

	/* Link it in */
	(*thisValue)->nextVariable = thisStackFrame->firstParameter;
	thisStackFrame->firstParameter = thisValue;
	
	return true;
}

/* Convert a variable to a string. Lock it and its string value before returning */
BOOL convertToString(HPCB Pcb, HVariable thisValue)
{
	/* First check that things are OK */
	
	short thisPosition;
	Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
	
	if (!thisValue) 
		return ScriptTR(Missing_Parameter, gScriptGlobals.stack->parsePosition,thisPosition);
	if ((*thisValue)->variableType == undefinedVariable)
		return ScriptTR(Uninitialized_Variable,gScriptGlobals.stack->parsePosition,thisPosition);
	HLock((Handle)thisValue);
		
	/* If it is a number then convert it. */
	
	if ((*thisValue)->variableType == numericVariable) {
		char Answer[256];
		Answer[0] = 0;
		sprintf(Answer, "%ld", (*thisValue)->Value.Number);
		if ( !((*thisValue)->Value.String = MyNewHandle(strlen(Answer)+1)) )
			return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		strcpy(*(*thisValue)->Value.String, Answer);
		(*thisValue)->variableType = stringVariable;
	}
	HLock((Handle)((*thisValue)->Value.String));
	return true;
}

/* Convert a variable to a number. Lock it before returning */
BOOL convertToInteger(HPCB Pcb, HVariable thisValue)
{
	/* First check that things are OK */

	short thisPosition;
	Parse_Get(Pcb, Scan_Position, &thisPosition, sizeof(thisPosition), false);
	
	if (!thisValue) 
		return ScriptTR(Missing_Parameter,gScriptGlobals.stack->parsePosition,thisPosition);
	if ((*thisValue)->variableType == undefinedVariable)
		return ScriptTR(Uninitialized_Variable,gScriptGlobals.stack->parsePosition,thisPosition);
	HLock((Handle)thisValue);
		
	/* If it is a string then convert it. */
	
	if ((*thisValue)->variableType == stringVariable) {
		long Answer;
		HLock((Handle)((*thisValue)->Value.String));
		c2pstr(*(*thisValue)->Value.String);
		StringToNum((StringPtr)(*(*thisValue)->Value.String), &Answer);
		DisposHandle((*thisValue)->Value.String);
		(*thisValue)->Value.Number = Answer;
		(*thisValue)->variableType = numericVariable;
	}
	return true;
}

BOOL MakeLocalVariable(PScript_Stack_Frame thisStackFrame, char *varName)
{
	HVariable thisVar;
	
	/* Check to see if this variable is already defined. */
	
	for (thisVar=thisStackFrame->localVariables; thisVar; 
			thisVar = (*thisVar)->nextVariable)
		if (0==strcmp(varName, (*thisVar)->variableName)) return true;
	
	if ( !(thisVar=NewVariable(varName, thisStackFrame->localVariables)) )
		return ScriptTR(Memory_Error,MyMemErr(),__LINE__);
	thisStackFrame->localVariables = thisVar;
	return true;
}

HVariable GetVariable(Script_Globals * TheGlobals, char *varName)
{
	HVariable thisVar=nil;
	register PScript_Stack_Frame thisFrame;
	
	/* look back through the stack for a local variable with this name. */
	
	for (thisFrame = TheGlobals->stack; thisFrame && !thisVar; thisFrame = thisFrame->nextFrame)
		for (thisVar=thisFrame->localVariables; thisVar; thisVar = (*thisVar)->nextVariable)
			if (0==strcmp(varName, (*thisVar)->variableName)) break;
	
	/* if we didn't find a local variable, look through the globals. */
	
	if (!thisVar)
		for (thisVar = TheGlobals->globalVariables; thisVar; thisVar = (*thisVar)->nextVariable)
			if (0==strcmp(varName, (*thisVar)->variableName)) break;
		
	/* If we didn't find a variable, then create a new global variable */
	
	if (!thisVar)
		if ( !(thisVar = NewVariable(varName, TheGlobals->globalVariables)) )
			ScriptTR(Memory_Error,MyMemErr(),__LINE__);
		else
			TheGlobals->globalVariables = thisVar;
			
	return thisVar;
}

HVariable NewVariable(char *varName, HVariable varList)
{
	HVariable thisVariable;
	if (thisVariable = (HVariable)MyNewHandle(sizeof(Variable)+strlen(varName)+1))
	{
		(*thisVariable)->nextVariable = varList;
		(*thisVariable)->variableType = undefinedVariable;
		strcpy((*thisVariable)->variableName, varName);
	}
	return thisVariable;
}

/* FreeStack disposes of a whole stack of stuff */
void FreeStack(PScript_Stack_Frame thisStack)
{
	while (thisStack) {
		PScript_Stack_Frame nextStack;
		DisposeVariables(thisStack->firstParameter);
		DisposeVariables(thisStack->localVariables);
		nextStack = thisStack->nextFrame;
		DisposPtr((Ptr)thisStack);
		thisStack = nextStack;
	}
}

/* DisposeVariable frees a linked list of variables */
void DisposeVariables(HVariable thisVar)
{
	while (thisVar) {
		HVariable nextVar;
		if ((*thisVar)->variableType == stringVariable)
			DisposHandle((*thisVar)->Value.String);
		nextVar = (*thisVar)->nextVariable;
		DisposHandle((Handle)thisVar);
		thisVar = nextVar;
	}
}

/* Report an error */
BOOL ScriptTR(UINT Script_Error,short Start,short End)
{
	Handle strnH;
	Str127 message,debugStr;
	strcpy((Ptr)message, "An error occurred while executing the login script. ");
	strnH = GetResource('STR#', 8800);
	if (strnH) {
		char *errString;
		if (Script_Error >= 0 && Script_Error < (*strnH)[0]*256 + (*strnH)[1]) {
			errString = *strnH+2;
			for (; Script_Error > 0;Script_Error--)
				errString += (*errString)+1;
			strncat((Ptr)message, errString+1, *errString);
		}
		else {
			Str127 strNum;
			NumToString(Script_Error,strNum);
			strcat((Ptr)message, (Ptr)strNum);
		}
		ReleaseResource(strnH);
		sprintf((Ptr)debugStr," {%d:%d}",Start,End);
		strcat((Ptr)message, (Ptr)debugStr);
		ErrorMessage((Ptr)message);
	}
	return false;
}
